--------------------------------------------------------------------------------
-- ReSynth Tool
--
-- Copyright 2011 Martin Bealby
--
-- Analog Oscillator Class (AnalogOsc)
--
-- Features:
--   - Pseudo-analog waveform synthesis
--   - Four different base waveforms (sin, pulse, sawtooth and noise)
--   - Waveform modification function, which works as follows
--     - sine/tri wave: sin (-63) -> triangle (+63)
--     - pulse wave   : pulse width modulation (-63 = 0%, +63 = 100%)
--     - sawtooth wave: no modification at present
--     - noise        : no modification at present
--   - Waveform modification LFO with both depth and rate (cycles) settings
--
-- Also note that the oscillator generates samples at 33489hz to give a waveform
-- length of 128 samples for a middle C note.  At this rate, there is no
-- advantage to making the samples 16 bit for these basic waveforms.
--
--------------------------------------------------------------------------------


--------------------------------------------------------------------------------
-- Enumerations
--------------------------------------------------------------------------------
ANALOGOSC_WAVEFORM_SINE     = 1
ANALOGOSC_WAVEFORM_PULSE    = 2
ANALOGOSC_WAVEFORM_SAWTOOTH = 3
ANALOGOSC_WAVEFORM_NOISE    = 4


--------------------------------------------------------------------------------
-- Preamble
------------------------------------------------------------------------------
class "AnalogOsc" (BaseOsc)


--------------------------------------------------------------------------------
-- Variables
--------------------------------------------------------------------------------
--[[
self.waveform                               (from enum above)
self.wavemod                                (modification amount: 0..63)
self.wavemod_lfo_depth                      (modification LFO depth: 0..31)
self.wavemod_lfo_rate                       (modification rate in cycles: 0..256)
]]--


--------------------------------------------------------------------------------
-- Functions
--------------------------------------------------------------------------------
function AnalogOsc:__init()
  -- Initialise oscillator to defaults
  BaseOsc.__init(self)
  self.waveform = ANALOGOSC_WAVEFORM_SINE
  self.wavemod = 0
  self.wavemod_lfo_depth = 0
  self.wavemod_lfo_rate = 0
end


function AnalogOsc:bind(sample)
  -- Binds the oscillator to a sample slot, sets parameters and renders
  BaseOsc.bind(self, sample)

  -- Linear interpolation for moar grit on pulse waves
  if self.smooth == true then
    sample.interpolation_mode = renoise.Sample.INTERPOLATE_LINEAR
  else
    sample.interpolation_mode = renoise.Sample.INTERPOLATE_NONE
  end
  
  -- init to 1 sample to regenerate on render call
  sample.sample_buffer:create_sample_data(33489,  -- freq (= 128*261.3hz)
                                          8,
                                          1,      -- channels
                                          32768)  -- frames (128 * 256)
end


function AnalogOsc:get_oscillator_type()
  -- Returns the oscillator type
  return "AnalogOsc"
end


function AnalogOsc:set_waveform(waveform)
  -- Sets the oscillator waveform ID
  assert(type(waveform) == "number")
  assert(waveform >= 1)
  assert(waveform <= 4)
  self.waveform = waveform
  self:render()
end


function AnalogOsc:get_waveform()
  -- Returns the current oscillator waveform ID
  return self.waveform
end


function AnalogOsc:set_wavemod(amount)
  -- Sets the waveform modification amount
  assert(type(amount) == "number")
  assert(amount >= 0)
  assert(amount <= 63)
  self.wavemod = amount
  self:render()
end


function AnalogOsc:get_wavemod()
  -- Returns the waveform modification amount
  return self.wavemod
end


function AnalogOsc:set_wavemod_lfo_depth(amount)
  -- Sets the waveform LFO modification amount
  assert(type(amount) == "number")
  assert(amount >= 0)
  assert(amount <= 127)
  self.wavemod_lfo_depth = amount
  self:render()
end


function AnalogOsc:get_wavemod_lfo_depth()
  -- Returns the waveform LFO modification amount
  return self.wavemod_lfo_depth
end


function AnalogOsc:set_wavemod_lfo_rate(cycles)
  -- Sets the waveform LFO modification rate
  assert(type(cycles) == "number")
  assert(cycles >= 0)
  assert(cycles <= 256)
  self.wavemod_lfo_rate = cycles
  self:render()
end


function AnalogOsc:get_wavemod_lfo_rate()
  -- Returns the waveform LFO modification rate
  return self.wavemod_lfo_rate
end


function AnalogOsc:set_smooth(smooth)
  -- Sets the oscillator smoothing toggle state
  assert(type(smooth) == "boolean")

  self.smooth = smooth
  if self.sample then
    if smooth == true then
      self.sample.interpolation_mode = renoise.Sample.INTERPOLATE_LINEAR
    else
      self.sample.interpolation_mode = renoise.Sample.INTERPOLATE_NONE
    end
  end
end


function AnalogOsc:render()
  -- The main sample synthesis function 
  
  if self.sample then
    -- We are bound, render

    local wt = {} -- local Lua waveform
    local pi = math.pi
    local sin = math.sin
    
    if self.waveform == ANALOGOSC_WAVEFORM_SINE then
      -- sine based oscillator   
      
      if self.wavemod_lfo_rate > 0 then
        if self.wavemod_lfo_depth > 0 then
          -- we have wavemod LFO enabled
          local this_cycle_wavemod = 0
          local user_mod = self.wavemod / 64   -- scaled from 0 to 1
          local lfo_amount = self.wavemod_lfo_depth / 32 -- scaled from 0 to 1
          local cycles = self.wavemod_lfo_rate
          
          for c = 1, cycles do
            this_cycle_wavemod = math.min(1.99,
                                          math.max(1,
                                                   user_mod +
                                                   (sin((c*pi)/cycles) * lfo_amount)   -- delta varies from -1 to +1
                                                   +1)) 
            for i = 1, 64 do
              table.insert(wt,  sin((i*pi)/64)^this_cycle_wavemod)
            end
            
            for i = 1, 64 do
              table.insert(wt,  -sin((i*pi)/64)^this_cycle_wavemod)
            end
          end
        else
          -- depth = 0, ignore and render as with no wavemod LFO
          local scaled_wavemod = (self.wavemod + 63)/128 + 1
          for i = 1, 64 do
            table.insert(wt,  sin((i*pi)/64)^scaled_wavemod)
          end
          
          for i = 1, 64 do
            table.insert(wt, -wt[i])
          end
        end
      else
        -- no wavemod LFO enabled
        local scaled_wavemod = (self.wavemod / 64) + 1
        for i = 1, 64 do
          table.insert(wt,  sin((i*pi)/64)^scaled_wavemod)
        end
        
        for i = 1, 64 do
          table.insert(wt, -wt[i])
        end
      end
      
      
    elseif self.waveform == ANALOGOSC_WAVEFORM_PULSE then
      -- pulse based oscillator
      
      if self.wavemod_lfo_rate > 0 then
        if self.wavemod_lfo_depth > 0 then
          -- we have wavemod LFO enabled

          local this_cycle_wavemod = 0
          local user_mod = self.wavemod
          local lfo_amount = self.wavemod_lfo_depth
          local cycles = self.wavemod_lfo_rate

          for c = 1, cycles do
            -- 2.1 bugfix (clamped to 0 < x < 64)
            this_cycle_wavemod = math.max(0,
                                          math.min(63,
                                                   math.floor(user_mod + (sin((c*pi)/cycles) * lfo_amount))))
          
            -- render each cycle
            for i = 1, 64 + this_cycle_wavemod do
              table.insert(wt, 1)
            end
            for i = 1, 64 - this_cycle_wavemod do
              table.insert(wt, -1)
            end
          end
        else
          -- depth = 0, ignore and render as with no wavemod LFO
          
          for i = 1, 64 + self.wavemod do
            table.insert(wt, 1)
          end
          for i = 1, 64 - self.wavemod do
            table.insert(wt, -1)
          end
        end
      else
        -- no wavemod LFO enabled
        for i = 1, 64 + self.wavemod do
          table.insert(wt, 1)
        end
        for i = 1, 64 - self.wavemod do
          table.insert(wt, -1)
        end
      end
      
    elseif self.waveform == ANALOGOSC_WAVEFORM_SAWTOOTH then
      -- sawtooth based oscillator
      -- no wavemod 
      for i = 1, 64 do
        table.insert(wt,  i*0.015625) -- (i *(1/64))
      end
        
      -- mirror and flip second half of waveform
      for i = 1, 64 do
        table.insert(wt, -wt[65-i])
      end
 
    elseif self.waveform == ANALOGOSC_WAVEFORM_NOISE then
      -- noise oscillator
      -- no wavemod
      math.randomseed(os.time())
      local random = math.random
      local sqrt = math.sqrt
      local cos = math.cos
      local r1
      local r2
      
      for i = 1, 32768 do
        table.insert(wt, sqrt(random()) * cos(2*pi*random()))
      end
   
 
    else
      -- shouldn't be here (invalid waveform)!
      assert(false,"Invalid waveform passed to AnalogOsc:render()")
    end
    
    -- At this point, table wt holds the entire sample in Luaspace 
    
    local sb = self.sample.sample_buffer

    -- 2.1 assert check
    assert(#wt < 32768+1, "Analog osc wavetable > 32768")
    
    self.sample.loop_mode = renoise.Sample.LOOP_MODE_FORWARD
    self.sample.loop_start = 1
    self.sample.loop_end = #wt
    self.sample.name = ""
    
    -- Dump to Renoise
    sb:prepare_sample_data_changes() 
    
    for i = 1,#wt do
      sb:set_sample_data(1, i, wt[i])
    end
    
    if #wt < 32768 then
      for i = #wt, 32768 do
        sb:set_sample_data(1, i, 0)
      end
    end
      
    sb:finalize_sample_data_changes()
  end
end


function AnalogOsc:get_settings()
  -- Returns all the analogue oscillator settings as a string
  -- AN1:waveform:wavemod:wavemod_lfo_depth:wavemod_lfo_rate:smoothing:transpose:fine_tune:volume:panning
  local str = "AN1:"
  str = str .. tostring(self.waveform) .. ":"
  str = str .. tostring(self.wavemod) .. ":"
  str = str .. tostring(self.wavemod_lfo_depth) .. ":"
  str = str .. tostring(self.wavemod_lfo_rate) .. ":"
  str = str .. tostring(self.smooth) .. ":"
  str = str .. tostring(self.transpose) .. ":"
  str = str .. tostring(self.fine_tune) .. ":"
  str = str .. tostring(math.lin2db(self.volume)) .. ":"
  str = str .. tostring(self.panning)
  return str
end


function AnalogOsc:set_settings(str)
  -- Sets all oscillator settings from the settings string supplied
  local settings_table = string_split(str, ":")
  assert(#settings_table == 10,
         "Wrong number of settings passed to set_settings, aborting.")
  assert(settings_table[1] == "AN1",
         "Not a valid AnalogOsc settings string passed to set_settings.")

  self.waveform = tonumber(settings_table[2])
  self.wavemod = tonumber(settings_table[3])
  self.wavemod_lfo_depth = tonumber(settings_table[4])
  self.wavemod_lfo_rate = tonumber(settings_table[5])
  if settings_table[6] == "true" then
    self.smooth = true
  else
    self.smooth = false
  end
  self.transpose = tonumber(settings_table[7])
  self.fine_tune = tonumber(settings_table[8])
  self.volume = math.db2lin(tonumber(settings_table[9]))
  self.panning = tonumber(settings_table[10])
end



