--------------------------------------------------------------------------------
-- ReSynth Tool
--
-- Copyright 2011 Martin Bealby
--
-- Base Oscillator Class (BaseOsc)
--
-- Features:
--   - Tranpose, finetune parameters
--   - Volume and panning parameters
--
--------------------------------------------------------------------------------



--------------------------------------------------------------------------------
-- Preamble
------------------------------------------------------------------------------
class "BaseOsc"


--------------------------------------------------------------------------------
-- Variables
--------------------------------------------------------------------------------
--[[
self.sample                                 (instrument sample reference)
self.smooth                                 (smooths the oscillator by interpolation)
self.transpose                              (sample transpose: -24..+24)
self.fine_tune                              (sample finetune: -63..+63)
self.volume                                 (sample volume: linear value)
self.panning                                (sample panning: -50 to +50)
]]--


--------------------------------------------------------------------------------
-- Functions
--------------------------------------------------------------------------------
function BaseOsc:__init()
  -- Initialise oscillator to defaults 
  self.sample = nil -- unbound
  self.smooth = true
  self.transpose = 0
  self.fine_tune = 0
  self.volume = math.db2lin(-12)
  self.panning = 0
end


function BaseOsc:bind(sample)
  -- Binds the oscillator to a sample slot, sets parameters
  assert(type(sample) == "Sample")
  self.sample = sample
  sample.transpose = self.transpose
  sample.fine_tune = self.fine_tune
  sample.volume = self.volume
  sample.panning = (self.panning + 50)/100
  if self.smooth == true then
    sample.interpolation_mode = renoise.Sample.INTERPOLATE_CUBIC
  else
    sample.interpolation_mode = renoise.Sample.INTERPOLATE_NONE
  end
end


function BaseOsc:unbind()
  -- Unbinds the oscillator from a sample slot
  self.sample = nil
end



function BaseOsc:set_transpose(semitones)
  -- Sets the oscillator transpose
  assert(type(semitones) == "number")
  assert(semitones >= -24)
  assert(semitones <= 24)
  self.transpose = semitones
  if self.sample then
    self.sample.transpose = semitones
  end
end


function BaseOsc:get_transpose()
  -- Returns the oscillator transpose
  return self.transpose
end


function BaseOsc:set_smooth(smooth)
  -- Sets the oscillator smoothing toggle state
  assert(type(smooth) == "boolean")

  self.smooth = smooth
  if self.sample then
    if smooth == true then
      self.sample.interpolation_mode = renoise.Sample.INTERPOLATE_LINEAR
    else
      self.sample.interpolation_mode = renoise.Sample.INTERPOLATE_NONE
    end
  end
end


function BaseOsc:get_smooth()
  -- Returns the oscillator smoothing toggle state
  return self.smooth
end


function BaseOsc:set_fine_tune(cents)
  -- Sets the oscillator fine tuning
  assert(type(cents) == "number")
  assert(cents >= -63)
  assert(cents <= 63)
  self.fine_tune = cents
  if self.sample then
    self.sample.fine_tune = cents
  end
end


function BaseOsc:get_fine_tune()
  -- Returns the oscillator fine tuning
  return self.fine_tune
end


function BaseOsc:set_volume(dB)
  -- Sets the oscillator volume
  assert(type(dB) == "number")
  assert(dB >= math.infdb)
  assert(dB <= 3)
  self.volume = math.db2lin(dB)
  if self.sample then
    self.sample.volume = self.volume
  end
end


function BaseOsc:get_volume()
  -- Returns the oscillator volume
  local v = math.lin2db(self.volume)
  if v > -36 then
    return math.min(3, v)  -- Wierd bug found by Tom de Rooy (asset db <= 3 fails with db = 3 in set_volume?)
  else
    return VOL_ABSMIN
  end
end



function BaseOsc:set_panning(pan)
  -- Sets the oscillator panning
  assert(type(pan) == "number")
  assert(pan >= -50)
  assert(pan <= 50)
  self.panning = pan
  if self.sample then
    self.sample.panning = (pan + 50)/100
  end
end


function BaseOsc:get_panning()
  -- Returns the oscillator panning
  return self.panning
end
