--------------------------------------------------------------------------------
-- ReSynth Tool
--
-- Copyright 2011 Martin Bealby
--
-- Main tool code
--------------------------------------------------------------------------------



--------------------------------------------------------------------------------
-- Constants
--------------------------------------------------------------------------------
ANALOG_OSC    = 1
WAVETABLE_OSC = 2
VOL_CUTOUT    = -36
VOL_ABSMIN    = -38


--------------------------------------------------------------------------------
-- Includes part 1
--------------------------------------------------------------------------------
require "support"
require "baseosc"
require "analogosc"
require "wavetableosc"
require "actions"
require "gui"
require "midimaps"

--------------------------------------------------------------------------------
-- Variables
--------------------------------------------------------------------------------
osc_an = {AnalogOsc(), AnalogOsc(), AnalogOsc(), AnalogOsc()}
osc_wt = {WavetableOsc(), WavetableOsc(), WavetableOsc(), WavetableOsc()}
osc_type = {ANALOG_OSC, ANALOG_OSC, ANALOG_OSC, ANALOG_OSC}
inst = nil


--------------------------------------------------------------------------------
-- Functions
--------------------------------------------------------------------------------
function new_resynth_instrument()
  -- Inserts a new instrument of 'resynth' type
  local rs = renoise.song()
  
  inst = rs:insert_instrument_at(#rs.instruments+1)
  rs.selected_instrument_index = #rs.instruments
  inst.name = "New Resynth Inst"
  
  -- generate 3 more sample slots for the 4 oscillators
  inst:insert_sample_at(1)
  inst:insert_sample_at(1)
  inst:insert_sample_at(1)
  
  -- create sample mappings
  inst:insert_sample_mapping(renoise.Instrument.LAYER_NOTE_ON,
                             1,       -- osc 1
                             48,      -- base @ C4
                             {0,119}, -- full key range
                             {0,127}) -- full vel range 
  inst:insert_sample_mapping(renoise.Instrument.LAYER_NOTE_ON,
                             2,       -- osc 2
                             48,      -- base @ C4
                             {0,119}, -- full key range
                             {0,127}) -- full vel range
  inst:insert_sample_mapping(renoise.Instrument.LAYER_NOTE_ON,
                             3,       -- osc 3
                             48,      -- base @ C4
                             {0,119}, -- full key range
                             {0,127}) -- full vel range 

  
  
  -- bind oscillators and set sample names
  for o = 1, 4 do
    if osc_type[o] == ANALOG_OSC then
      osc_an[o]:bind(inst.samples[o])
      osc_an[o]:render()
      inst.samples[o].name = osc_an[o]:get_settings()
    elseif osc_type[o] == WAVETABLE_OSC then
      osc_wt[o]:bind(inst.samples[o])
      inst.samples[o].name = osc_wt[o]:get_settings()
    end
  end
  
  -- Show GUI
  display()
  
  -- Reset
  global_reset()
end


function open_resynth_instrument(new_inst)
  -- Open an existing ReSynth instrument
  
  -- Sense check - sample count
  if #new_inst.samples ~=4 then
    local inst_name = new_inst.name
    if inst_name ~= "" then
      renoise.app():show_warning(inst_name ..
                               " is an invalid or corrupt ReSynth2 instrument.")
    else
      renoise.app():show_warning("Invalid or corrupt ReSynth2 instrument.")
    end
    return
  end
  
  -- Sense check - sample ID headers
  for i = 1, 4 do
    local samp_name_start = string.sub(new_inst.samples[i].name, 1, 3)
    if samp_name_start ~= "AN1" then
      if samp_name_start ~= "WT1" then
        local inst_name = new_inst.name
        if inst_name ~= "" then
          renoise.app():show_warning(inst_name ..
                                   " is an invalid or corrupt ReSynth2 instrument.")
        else
          renoise.app():show_warning("Invalid or corrupt ReSynth2 instrument.")
        end
        return
      end
    end
  end
 
  -- Seems ok, assign
  inst = new_inst
  
  -- Bind and initialise oscillators
  for o = 1, 4 do
    local settings = inst.samples[o].name
    local osc_id = string.sub(settings, 1, 3)
    if osc_id == "AN1" then
      osc_type[o] = ANALOG_OSC
    else
      osc_type[o] = WAVETABLE_OSC
    end 
    if osc_type[o] == ANALOG_OSC then
      osc_an[o]:set_settings(settings)
      osc_an[o]:bind(inst.samples[o])
      osc_an[o]:set_settings(settings)
      osc_an[o]:render()
      inst.samples[o].name = osc_an[o]:get_settings()
      
      -- TODO: hacky work around for common controls
      osc_wt[o]:set_smooth(osc_an[o]:get_smooth())
      osc_wt[o]:set_transpose(osc_an[o]:get_transpose())
      osc_wt[o]:set_fine_tune(osc_an[o]:get_fine_tune())
      osc_wt[o]:set_volume(osc_an[o]:get_volume())
      osc_wt[o]:set_panning(osc_an[o]:get_panning())

    elseif osc_type[o] == WAVETABLE_OSC then
      osc_wt[o]:set_settings(settings)
      osc_wt[o]:bind(inst.samples[o])
      osc_wt[o]:set_settings(settings)
      inst.samples[o].name = osc_wt[o]:get_settings()
      
      -- TODO: hacky work around for common controls
      osc_an[o]:set_smooth(osc_wt[o]:get_smooth())
      osc_an[o]:set_transpose(osc_wt[o]:get_transpose())
      osc_an[o]:set_fine_tune(osc_wt[o]:get_fine_tune())
      osc_an[o]:set_volume(osc_wt[o]:get_volume())
      osc_an[o]:set_panning(osc_wt[o]:get_panning())
    end
  end
  
  -- Show the GUI
  display()
end


function open_waveforms_folder()
  -- Open the tools' waveforms folder
  
  local folder = os.currentdir().."/waveforms"
  if io.exists(folder) then
    renoise.app():open_path(folder)
  else
    renoise.app():show_error("Waveforms folder missing.")
  end
end

--------------------------------------------------------------------------------
-- Menu Integration
--------------------------------------------------------------------------------
renoise.tool():add_menu_entry {
  name = "Instrument Box:Open as Resynth Instrument",
  invoke = function() open_resynth_instrument(renoise.song().selected_instrument) end
}

renoise.tool():add_menu_entry {
  name = "Instrument Box:New Resynth Instrument",
  invoke = function() new_resynth_instrument() end
}


renoise.tool():add_menu_entry {
  name = "Main Menu:Tools:ReSynth:Open current instrument as a Resynth Instrument",
  invoke = function() open_resynth_instrument(renoise.song().selected_instrument) end
}

renoise.tool():add_menu_entry {
  name = "Main Menu:Tools:ReSynth:New Resynth Instrument",
  invoke = function() new_resynth_instrument() end
}


renoise.tool():add_menu_entry {
  name = "---Main Menu:Tools:ReSynth:Open Waveforms Folder",
  invoke = function() open_waveforms_folder() end
}



--------------------------------------------------------------------------------
-- Keybindings
--------------------------------------------------------------------------------
renoise.tool():add_keybinding {
  name = "Global:ReSynth:New Resynth Instrument",
  invoke = function() new_resynth_instrument() end
}


renoise.tool():add_keybinding {
  name = "Global:ReSynth:Open as Resynth Instrument",
  invoke = function() open_resynth_instrument(renoise.song().selected_instrument) end
}

renoise.tool():add_keybinding {
  name = "Global:ReSynth:Open Waveforms Folder",
  invoke = function() open_waveforms_folder() end
}
