--------------------------------------------------------------------------------
-- ReSynth Tool
--
-- Copyright 2011 Martin Bealby
--
-- Wavetable Oscillator Class (WavetableOsc)
--
-- Features:
--   - Multiple banks of wavetables
--
-- The wavetable oscillator will search for banks within a waveforms folder
-- inside the tool:
-- e.g. __tool_name__/waveforms/bankname/waveform.wav
--
-- All waveforms must be tuned to middle C and be in wav format
--
-- Waveforms will only be loaded from the load_waveform() function
--
--------------------------------------------------------------------------------



--------------------------------------------------------------------------------
-- Preamble
------------------------------------------------------------------------------
class "WavetableOsc" (BaseOsc)



--------------------------------------------------------------------------------
-- Constants
--------------------------------------------------------------------------------
AUTOLOOP_THRESHOLD = 256 --samples


--------------------------------------------------------------------------------
-- Variables
--------------------------------------------------------------------------------
--[[
self.wavebank                               (current waveform bank reference)
self.waveform                               (current waveform reference)
]]--


--------------------------------------------------------------------------------
-- Functions
--------------------------------------------------------------------------------
function WavetableOsc:__init()
  -- Initialise oscillator to defaults 
  BaseOsc.__init(self)
  self.wavebank = ""
  self.waveform = ""
end


function WavetableOsc:load_waveform(bank, waveform)
  -- Sets the oscillator waveform ID
  assert(type(bank) == "string")
  assert(type(waveform) == "string")
  assert(io.exists(os.currentdir().."/waveforms"))
  assert(io.exists(os.currentdir().."/waveforms/"..bank))
  assert(io.exists(os.currentdir().."/waveforms/"..bank.."/"..waveform))
  
  -- Load the sample only if we are already bound
  if self.sample then
    self.waveform = waveform
    self.wavebank = bank
    self.sample.sample_buffer:load_from(os.currentdir() ..
                                        "/waveforms/" ..
                                        bank ..
                                        "/" ..
                                        waveform)
    self.sample.name = ""
  end
  
  -- Enable looping for short samples
  if self.sample.sample_buffer.number_of_frames <= AUTOLOOP_THRESHOLD then
    self.sample.loop_start = 1
    self.sample.loop_end = self.sample.sample_buffer.number_of_frames
    self.sample.loop_mode = renoise.Sample.LOOP_MODE_FORWARD
  end
end


function WavetableOsc:get_oscillator_type()
  -- Returns the oscillator type
  return "WavetableOsc"
end


function WavetableOsc:get_waveform()
  -- Returns the current oscillator waveform ID
  return self.waveform
end


function WavetableOsc:get_wavebank()
  -- Returns the current oscillator waveform ID
  return table.find(self:get_bank_list(), self.wavebank)
end


function WavetableOsc:get_bank_list()
  -- Returns a table of available wavebanks
  assert(io.exists(os.currentdir().."/waveforms"))
  return os.dirnames(os.currentdir().."/waveforms")
end


function WavetableOsc:get_waveform_list(bank)
  -- Returns a table of available waveforms within the specified bank
  assert(type(bank) == "string")
  if io.exists(os.currentdir().."/waveforms/"..bank) then
    return (os.filenames(os.currentdir().."/waveforms/"..bank,
            {"*.wav"}))
  else
    return {}
  end
end


function WavetableOsc:get_settings()
  -- Returns all the wavetable oscillator settings as a string
  -- WT1:<bank>:<waveform>:smoothing:transpose:fine_tune:volume:panning
  local str = "WT1:"
  str = str .. self.wavebank .. ":"
  str = str .. self.waveform .. ":"
  str = str .. tostring(self.smooth) .. ":"
  str = str .. tostring(self.transpose) .. ":"
  str = str .. tostring(self.fine_tune) .. ":"
  str = str .. tostring(math.lin2db(self.volume)) .. ":"
  str = str .. tostring(self.panning)
  return str
end


function WavetableOsc:set_settings(str)
  -- Sets all oscillator settings from the settings string supplied
  local settings_table = string_split(str, ":")
  assert(#settings_table == 8,
         "Wrong number of settings passed to set_settings, aborting.")
  assert(settings_table[1] == "WT1",
         "Not a valid WavetableOsc settings string passed to set_settings.")

  self.wavebank = settings_table[2]
  self.waveform =settings_table[3]
  if settings_table[4] == "true" then
    self.smooth = true
  else
    self.smooth = false
  end
  self.transpose = tonumber(settings_table[5])
  self.fine_tune = tonumber(settings_table[6])
  self.volume = math.db2lin(tonumber(settings_table[7]))
  self:set_panning(tonumber(settings_table[8]))
end
